from django.contrib.auth.models import AnonymousUser
from django.core.handlers.wsgi import WSGIRequest
from django.test import Client, RequestFactory, TestCase

import graphene
from graphene_django.settings import graphene_settings

from .views import GraphQLView


class SchemaRequestFactory(RequestFactory):

    def __init__(self, **defaults):
        super().__init__(**defaults)
        self._schema = graphene_settings.SCHEMA

    def schema(self, **kwargs):
        self._schema = graphene.Schema(**kwargs)

    def execute(self, query, **options):
        return self._schema.execute(query, **options)


class SchemaClient(SchemaRequestFactory, Client):

    def __init__(self, **defaults):
        super().__init__(**defaults)
        self._credentials = {}
        self._user = AnonymousUser()

    def request(self, **request):
        request = WSGIRequest(self._base_environ(**request))
        request.user = self._user
        return request

    def authenticate(self, user):
        self._user = user

    def execute(self, query, variables=None, **extra):
        extra.update(self._credentials)
        context = self.post('/', **extra)

        result = super().execute(
            query,
            context_value=context,
            variable_values=variables,
        )
        if result.errors is not None:
            result.errors = [
                GraphQLView.format_error(error)
                for error in result.errors
            ]
        return result


class SchemaTestCase(TestCase):
    client_class = SchemaClient



from functools import wraps

from graphql.execution.execute import GraphQLResolveInfo

from . import exceptions

__all__ = [
    'user_passes_test',
    'login_required',
    'staff_member_required',
    'superuser_required',
    'permission_required',
]


def context(f):
    def decorator(func):
        def wrapper(*args, **kwargs):
            info = next(
                arg for arg in args
                if isinstance(arg, GraphQLResolveInfo)
            )
            return func(info.context, *args, **kwargs)
        return wrapper
    return decorator


def user_passes_test(test_func, exc=exceptions.PermissionDenied):
    def decorator(f):
        @wraps(f)
        @context(f)
        def wrapper(context, *args, **kwargs):
            if test_func(context.user):
                return f(*args, **kwargs)
            raise exc
        return wrapper
    return decorator


login_required = user_passes_test(lambda u: u.is_authenticated)
staff_member_required = user_passes_test(lambda u: u.is_active and u.is_staff)
superuser_required = user_passes_test(lambda u: u.is_superuser)


def permission_required(perm):
    def check_perms(user):
        if isinstance(perm, str):
            perms = (perm,)
        else:
            perms = perm

        if user.has_perms(perms):
            return True
        return False
    return user_passes_test(check_perms)



import traceback
from calendar import timegm
from datetime import datetime

from django.conf import settings
from django.utils.translation import gettext as _

from graphene_django.views import GraphQLView as BaseGraphQLView
from graphql.error import GraphQLError

from . import exceptions
from .settings import extensions_settings


def show_error_message(error):
    return settings.DEBUG or isinstance(error, (
        GraphQLError,
        exceptions.GraphQLError,
    ))


class GraphQLView(BaseGraphQLView):

    @staticmethod
    def format_error(error):
        formatted = BaseGraphQLView.format_error(error)

        if isinstance(error, GraphQLError) and\
                error.original_error is not None:

            error = error.original_error

        if not extensions_settings.SHOW_ERROR_MESSAGE_HANDLER(error):
            formatted['message'] = _('Internal server error')

        extensions = {
            'type': error.__class__.__name__,
            'code': getattr(error, 'code', 'error'),
            'timestamp': timegm(datetime.utcnow().utctimetuple()),
            **formatted.get('extensions', {}),
        }

        if hasattr(error, 'error_data'):
            extensions['data'] = error.error_data

        if error.__traceback__ is not None:
            info = error.__traceback__.tb_frame.f_locals.get('info')

            if info is not None:
                extensions['operation'] = info.operation.operation.name

            if settings.DEBUG:
                extensions['trace'] = traceback.format_list(
                    traceback.extract_tb(error.__traceback__),
                )

        formatted['extensions'] = extensions
        return formatted



from django.conf import settings
from django.test.signals import setting_changed
from django.utils.module_loading import import_string

DEFAULTS = {
    'SHOW_ERROR_MESSAGE_HANDLER':
    'graphql_extensions.views.show_error_message',
}

IMPORT_STRINGS = (
    'SHOW_ERROR_MESSAGE_HANDLER',
)


def perform_import(value, setting_name):
    if isinstance(value, str):
        return import_from_string(value, setting_name)
    if isinstance(value, (list, tuple)):
        return [import_from_string(item, setting_name) for item in value]
    return value


def import_from_string(value, setting_name):
    try:
        return import_string(value)
    except ImportError as e:
        msg = (
            f'Could not import `{value}` for EXTENSIONS setting'
            f'`{setting_name}`. {e.__class__.__name__}: {e}.'
        )
        raise ImportError(msg)


class ExtensionsSettings(object):

    def __init__(self, defaults, import_strings):
        self.defaults = defaults
        self.import_strings = import_strings
        self._cached_attrs = set()

    def __getattr__(self, attr):
        if attr not in self.defaults:
            raise AttributeError(f'Invalid setting: `{attr}`')

        value = self.user_settings.get(attr, self.defaults[attr])

        if attr in self.import_strings:
            value = perform_import(value, attr)

        self._cached_attrs.add(attr)
        setattr(self, attr, value)
        return value

    @property
    def user_settings(self):
        if not hasattr(self, '_user_settings'):
            self._user_settings = getattr(settings, 'GRAPHQL_EXTENSIONS', {})
        return self._user_settings

    def reload(self):
        for attr in self._cached_attrs:
            delattr(self, attr)

        self._cached_attrs.clear()

        if hasattr(self, '_user_settings'):
            delattr(self, '_user_settings')


def reload_settings(*args, **kwargs):
    setting = kwargs['setting']

    if setting == 'GRAPHQL_EXTENSIONS':
        extensions_settings.reload()


setting_changed.connect(reload_settings)

extensions_settings = ExtensionsSettings(DEFAULTS, IMPORT_STRINGS)



import datetime

from django.core.exceptions import ValidationError
from django.core.validators import validate_email

from graphene.types.scalars import Scalar
from graphql.language import ast

from . import exceptions


class Timestamp(Scalar):

    @staticmethod
    def serialize(ts):
        return int(ts.timestamp())

    @staticmethod
    def parse_literal(cls, node):
        if isinstance(node, ast.IntValueNode):
            return cls.parse_value(node.value)

    @staticmethod
    def parse_value(value):
        return datetime.datetime.fromtimestamp(value)


class Email(Scalar):

    @staticmethod
    def serialize(email):
        return str(email)

    @staticmethod
    def parse_literal(cls, node):
        if isinstance(node, ast.StringValueNode):
            return cls.parse_value(node.value)

    @staticmethod
    def parse_value(value):
        try:
            validate_email(value)
        except ValidationError as e:
            raise exceptions.ValidationError(str(e.args[0]))
        return value



CHOICES_LOOKUPS = ['exact', 'in']

TEXT_LOOKUPS = CHOICES_LOOKUPS + [
    'iexact', 'contains', 'icontains', 'startswith',
    'istartswith', 'endswith', 'iendswith', 'regex', 'iregex',
]

RANGE_LOOKUPS = ['exact', 'gt', 'gte', 'lt', 'lte']

DATE_UNITS = ['year', 'month', 'week_day', 'day']

DATE_LOOKUPS = RANGE_LOOKUPS + DATE_UNITS

DATETIME_LOOKUPS = RANGE_LOOKUPS + DATE_UNITS + ['hour'] + \
    ['date__' + lookup for lookup in RANGE_LOOKUPS]
